﻿using System.IO;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Markup;
using Exilebuddy;
using log4net;
using Loki.Bot;
using System;
using Loki.Common;
using Loki.Game;

namespace Scheduler
{
    internal class Scheduler : IPlugin
    {
        private static readonly ILog Log = Logger.GetLoggerInstanceForType();

        private bool _enabled;

        private UserControl root;

        #region Implementation of IPlugin

        /// <summary> The name of the plugin. </summary>
        public string Name
        {
            get { return "Scheduler"; }
        }

        /// <summary> The description of the plugin. </summary>
        public string Description
        {
            get { return "A plugin that allows for some simple scheduling to control how long the bot runs."; }
        }

        /// <summary>The author of the plugin.</summary>
        public string Author
        {
            get { return "Bossland GmbH"; }
        }

        public string Version
        {
            get
            {
                return "0.0.1.1";
            }
        }


        /// <summary>Initializes this plugin.</summary>
        public void Initialize()
        {
            Log.DebugFormat("[Scheduler] Initialize");

            SchedulerSettings.Instance.ProgramRuntimeStopwatch.Start();

            BotManager.OnBotChanged += BotManagerOnOnBotChanged;

			//UriBuilder uri = new UriBuilder(Assembly.GetExecutingAssembly().CodeBase);
			//string path = Uri.UnescapeDataString(uri.Path);
			//Console.Write(path);
        }
        #region Implementation of ILogic

        /// <summary>
        /// Coroutine logic to execute.
        /// </summary>
        /// <param name="type">The requested type of logic to execute.</param>
        /// <param name="param">Data sent to the object from the caller.</param>
        /// <returns>true if logic was executed to handle this type and false otherwise.</returns>
        public async Task<bool> Logic(string type, params dynamic[] param)
        {
            return false;
        }

        /// <summary>
        /// Non-coroutine logic to execute.
        /// </summary>
        /// <param name="name">The name of the logic to invoke.</param>
        /// <param name="param">The data passed to the logic.</param>
        /// <returns>Data from the executed logic.</returns>
        public object Execute(string name, params dynamic[] param)
        {
            return null;
        }

        #endregion
        public void Deinitialize()
        {
            Log.DebugFormat("[Scheduler] Deinitialize");
        }
        /// <summary> The plugin start callback. Do any initialization here. </summary>
        public void Start()
        {
            Log.DebugFormat("[Scheduler] Start");

            SchedulerSettings.Instance.BotIdleStopwatch.Reset();
            SchedulerSettings.Instance.WaitingForIdle = false;

            SchedulerSettings.Instance.BotRuntimeStopwatch.Start();

	        var r = new Random();

	        if (SchedulerSettings.Instance.BotTimeLimitMax < SchedulerSettings.Instance.BotTimeLimitMin)
	        {
		        SchedulerSettings.Instance.BotTimeLimitMax = SchedulerSettings.Instance.BotTimeLimitMin;
	        }
			SchedulerSettings.Instance.BotTimeLimit = TimeSpan.FromMilliseconds(SchedulerSettings.Instance.BotTimeLimitMin.TotalMilliseconds +
	                                                                     r.Next(0,
		                                                                     1 +
		                                                                     (int)
			                                                                     (SchedulerSettings.Instance.BotTimeLimitMax
																					 .TotalMilliseconds -
			                                                                      SchedulerSettings.Instance.BotTimeLimitMin
																					  .TotalMilliseconds)));

			if (SchedulerSettings.Instance.BotIdleTimeLimitMax < SchedulerSettings.Instance.BotIdleTimeLimitMin)
			{
				SchedulerSettings.Instance.BotIdleTimeLimitMax = SchedulerSettings.Instance.BotIdleTimeLimitMin;
			}

	        SchedulerSettings.Instance.BotIdleTimeLimit =
		        TimeSpan.FromMilliseconds(SchedulerSettings.Instance.BotIdleTimeLimitMin.TotalMilliseconds +
		                                  r.Next(0,
			                                  1 +
			                                  (int)
				                                  (SchedulerSettings.Instance.BotIdleTimeLimitMax.TotalMilliseconds -
				                                   SchedulerSettings.Instance.BotIdleTimeLimitMin.TotalMilliseconds)));

			Log.InfoFormat("[Scheduler] BotTimeLimit: {0}", SchedulerSettings.Instance.BotTimeLimit);
			Log.InfoFormat("[Scheduler] BotIdleTimeLimit: {0}", SchedulerSettings.Instance.BotIdleTimeLimit);
        }

        /// <summary> The plugin tick callback. Do any update logic here. </summary>
        public void Tick()
        {
        }

        /// <summary> The plugin stop callback. Do any pre-dispose cleanup here. </summary>
        public void Stop()
        {
            Log.DebugFormat("[Scheduler] Stop");

            SchedulerSettings.Instance.BotRuntimeStopwatch.Stop();
            SchedulerSettings.Instance.BotRuntimeStopwatch.Reset();

	        if (_executeStopLogic)
	        {
		        _executeStopLogic = false;

		        // We need to be in game first!
		        if (LokiPoe.IsInGame)
		        {
			        var ltcserr = LokiPoe.EscapeState.LogoutToTitleScreen();
			        if (ltcserr != LokiPoe.EscapeState.LogoutError.None)
			        {
				        Log.ErrorFormat("[Scheduler] LogoutToTitleScreen returned {0}.", ltcserr);
			        }
		        }
	        }
        }

        public JsonSettings Settings
        {
            get { return SchedulerSettings.Instance; }
        }

        /// <summary> The plugin's settings control. This will be added to the Exilebuddy Settings tab.</summary>
        public UserControl Control
        {
            get
            {
                using (var fs = new FileStream(@"3rdParty\Scheduler\SettingsGui.xaml", FileMode.Open))
                {
                    var root = (UserControl) XamlReader.Load(fs);

                    // Your settings binding here.

                    if (
                        !TimePicker.SetupSelectedTimeBinding(root, "ProgramRuntimeTimePicker",
                            "ProgramRuntime", BindingMode.OneWay, SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupCheckBoxBinding failed for 'ProgramRuntimeTimePicker'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !TimePicker.SetupSelectedTimeBinding(root, "BotIdleTimeTimePicker",
                            "BotIdleTime", BindingMode.OneWay, SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupCheckBoxBinding failed for 'BotIdleTimeTimePicker'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !TimePicker.SetupSelectedTimeBinding(root, "BotRuntimeTimePicker",
                            "BotRuntime", BindingMode.OneWay, SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupCheckBoxBinding failed for 'BotRuntimeTimePicker'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !TimePicker.SetupSelectedTimeBinding(root, "BotIdleTimeLimitMinTimePicker",
                            "BotIdleTimeLimitMin", BindingMode.TwoWay, SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
							"[SettingsControl] SetupCheckBoxBinding failed for 'BotIdleTimeLimitMinTimePicker'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

					if (
						!TimePicker.SetupSelectedTimeBinding(root, "BotIdleTimeLimitMaxTimePicker",
							"BotIdleTimeLimitMax", BindingMode.TwoWay, SchedulerSettings.Instance))
					{
						Log.DebugFormat(
							"[SettingsControl] SetupCheckBoxBinding failed for 'BotIdleTimeLimitMaxTimePicker'.");
						throw new Exception("The SettingsControl could not be created.");
					}

					if (
						!TimePicker.SetupSelectedTimeBinding(root, "BotTimeLimitMinTimePicker",
							"BotTimeLimitMin", BindingMode.TwoWay, SchedulerSettings.Instance))
					{
						Log.DebugFormat(
							"[SettingsControl] SetupCheckBoxBinding failed for 'BotTimeLimitMinTimePicker'.");
						throw new Exception("The SettingsControl could not be created.");
					}
                    if (
						!TimePicker.SetupSelectedTimeBinding(root, "BotTimeLimitMaxTimePicker",
                            "BotTimeLimitMax", BindingMode.TwoWay, SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
							"[SettingsControl] SetupCheckBoxBinding failed for 'BotTimeLimitMaxTimePicker'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupCheckBoxBinding(root, "StartBotAfterLimitHitCheckBox",
                        "StartBotAfterLimitHit", BindingMode.TwoWay,
                        SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'StartBotAfterLimitHitCheckBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupCheckBoxBinding(root, "LimitBotTimeCheckBox",
                        "LimitBotTime", BindingMode.TwoWay,
                        SchedulerSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'LimitBotTimeCheckBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    // Your settings event handlers here.

                    return root;
                }
            }
        }


        /// <summary> The plugin is being enabled.</summary>
        public void Enable()
        {
            Log.DebugFormat("[Scheduler] Enable");
            _enabled = true;

            LokiPoe.OnGuiTick += LokiPoeOnOnGuiTick;
        }

        /// <summary> The plugin is being disabled.</summary>
        public void Disable()
        {
            Log.DebugFormat("[Scheduler] Disable");
            _enabled = false;

            LokiPoe.OnGuiTick -= LokiPoeOnOnGuiTick;

            Log.DebugFormat("[OnDisable] Schedule settings reseting due to the plugin being disabled.");

            SchedulerSettings.Instance.BotRuntimeStopwatch.Reset();
            SchedulerSettings.Instance.BotIdleStopwatch.Reset();
            SchedulerSettings.Instance.WaitingForIdle = false;
        }

        #endregion

        #region Implementation of IDisposable

        /// <summary> </summary>
        public void Dispose()
        {
            SchedulerSettings.Instance.ProgramRuntimeStopwatch.Stop();
        }

        #endregion

        #region Override of Object

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return Name + ": " + Description;
        }

        #endregion

	    private bool _executeStopLogic = false;

        private void LokiPoeOnOnGuiTick(object sender, GuiTickEventArgs guiTickEventArgs)
        {
            // Doesn't actually set the value, triggers an update.
            SchedulerSettings.Instance.ProgramRuntime = TimeSpan.MaxValue;
            SchedulerSettings.Instance.BotRuntime = TimeSpan.MaxValue;
            SchedulerSettings.Instance.BotIdleTime = TimeSpan.MaxValue;

            // If we're waiting for the idle timer to expire.
            if (SchedulerSettings.Instance.WaitingForIdle)
            {
                // Check to see if we are restarting the bot.
                if (!SchedulerSettings.Instance.StartBotAfterLimitHit)
                    return;

                // Check to see if the desired elapsed time has passed yet.
                if (SchedulerSettings.Instance.BotIdleStopwatch.Elapsed < SchedulerSettings.Instance.BotIdleTimeLimit)
                    return;

                Log.DebugFormat("[LokiPoeOnOnGuiTick] The bot idle timer has expired. Now starting the bot!");

                // At this point, we're ready to restart the bot, so cleanup state first.
                SchedulerSettings.Instance.WaitingForIdle = false;
                SchedulerSettings.Instance.BotIdleStopwatch.Reset();

                SchedulerSettings.Instance.BotRuntimeStopwatch.Reset();

                BotManager.Start();

                // All done!
                return;
            }

            // If we need to limit the botting time.
            if (SchedulerSettings.Instance.LimitBotTime)
            {
                // Check to see if the desired elapsed time has passed yet.
                if (SchedulerSettings.Instance.BotRuntimeStopwatch.Elapsed < SchedulerSettings.Instance.BotTimeLimit)
                    return;

                if (!BotManager.IsRunning)
                {
                    return;
                }

                Log.DebugFormat(
                    "[LokiPoeOnOnGuiTick] The bot run timer has been reached. Now requesting the bot stop.");

				_executeStopLogic = true;

				BotManager.Stop();

				SchedulerSettings.Instance.WaitingForIdle = true;
				SchedulerSettings.Instance.BotIdleStopwatch.Restart();
            }
        }

        private void BotManagerOnOnBotChanged(object sender, BotChangedEventArgs botChangedEventArgs)
        {
            Log.DebugFormat("[BotManagerOnOnBotChanged] Schedule settings reseting due to a bot change.");

            SchedulerSettings.Instance.BotRuntimeStopwatch.Reset();
            SchedulerSettings.Instance.BotIdleStopwatch.Reset();
            SchedulerSettings.Instance.WaitingForIdle = false;
        }
    }
}
